/*

   befdeb.c
   Befunge Debugger
     - v1.0, Sept 1993, Chris Pressey
     - v1.01, Feb 1996, Chris Pressey

     - WIN32 and interactive debugging added by Kevin Vigor, April 1997

   (c)1993, 1995, 1996, Chris Pressey, Cat's-Eye Technologies
   cats-eye@cats-eye.com
   http://www.cats-eye.com/cet/soft/lang/befunge/

   usage :

   befdeb <befunge-source> <input-file> <output-file> <stack-file>

   compilation :

   Amiga : dcc befdeb.c -o befdeb (amiga/befdeb)
   Un*x  : gcc befdeb.c -o befdeb
   MS-DOS: used Borland C++ v3.1. (msdos/befdeb)

   see bef.c for v1.01 bug fix notes.  -o workaround is not
   present in debugger.

   Notes for the DOS version:
     Load ANSI.SYS or compatible ANSI driver before using this.
     define DOS_DELAY to any number suitable on your system if the
     included BEFDEB.EXE is too fast.
     To avoid including output to a file you can use NUL to use the
     DOS null device, or use NIL to suppress input (only!) from BEFDEB.

 */

/********************************************************* #INCLUDE'S */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

#if __BORLANDC__

   #include <dos.h>
   #define DELAY(x) delay(x)

#elif defined(WIN32)

   #include <windows.h>
   #include <conio.h>

   #define DELAY(x) Sleep(x)

   /* INTERACTIVE makes life much, much cooler, but will only work for WIN32 */
   #define INTERACTIVE

#endif

#ifdef DELAY
#define DELAY_MS 250 /* 250 millisecond pause */
#endif

/********************************************************** #DEFINE'S */

#define LINEWIDTH  79           /* 0 .. 78 */
#define PAGEHEIGHT 22           /* 0 .. 21 */

#define cur pg[y * LINEWIDTH + x]
#define shrink(s) s[strlen(s)-1]=0

/********************************************************* STRUCTURES */

struct stack                    /* stack structure, for values on stack */
  {
    signed long val;
    struct stack *next;
  }
 *head;                         /* head of stack */

/*************************************************** GLOBAL VARIABLES */

char pg[LINEWIDTH * PAGEHEIGHT]; /* befunge 'page' of source */
int x = 0, y = 0;                /* x and y of the PC */
int dx = 1, dy = 0;              /* direction of the PC */
int stringmode = 0;              /* flag : are we in string mode? */
int noin = 0;                    /* flag : no input file */

#ifdef WIN32

HANDLE  outputConsole;      /* Required for console output functions. */

#endif

#ifdef INTERACTIVE

int animate = 0;            /* Are we currently animating?? */
int go = 0;                 /* Are we currently going full pace?? */
int breakPointActive = 0;   /* Is there an active breakpoint? */

int oldX = 0, oldY = 0;     /* Holds X:Y from previous instruction */
int breakX = 0, breakY = 0; /* Breakpoint address. */

/* Colors used in INTERACTIVE mode. */
#define PC_COLOR    (BACKGROUND_RED | BACKGROUND_BLUE | BACKGROUND_GREEN)
#define NORMAL_COLOR (FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_GREEN)
#define BREAKPOINT_COLOR (NORMAL_COLOR | BACKGROUND_BLUE)

#endif

/********************************************************* PROTOTYPES */

void push (signed long val);
signed long pop (void);

void setConsolePosition(int x, int y);
void clearConsole(void);

#ifdef INTERACTIVE
void printStack(void);
void printPC(int x, int y);
void setupInteractiveConsole(void);
void debugInput(int, int);
void debugHighlightChar(int x, int y, WORD attribute);
void debuggerSay(char *msg);
#endif

/******************************************************* MAIN PROGRAM */

void main (argc, argv)
     int argc;
     char **argv;
{
  FILE *f;
  FILE *fi;
  FILE *fo;
  FILE *fs;

  srand (time (0));

  printf ("Befunge-93 Debugger Implementation v1.01\n");

#ifdef WIN32
    /* Get console info for WIN32. */
    outputConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    if (outputConsole == INVALID_HANDLE_VALUE)
    {
        puts("Win32: FATAL: cannot get output console handle.");
        exit(-1);
    }
#endif

  if (argc < 5)
    {
      printf ("USAGE : %s <befunge-source> <input-file> <output-file> <stack-file>\n", argv[0]);
      exit (0);
    }
  if ((f = fopen (argv[1], "r")) != NULL )            /*** Input Phase */
    {
      int x = 0, y = 0;
      while (!feof (f))
	{
	  cur = fgetc (f);
	  if (cur == '\n')
	    {
	      cur = ' ';
	      x = 0;
	      y++;
	      if (y >= PAGEHEIGHT)
		break;
	    }
	  else
	    {
	      x++;
	      if (x >= LINEWIDTH)
		{
		  x = 0;
		  y++;
		  if (y >= PAGEHEIGHT)
		    break;
		}
	    }
	}
      fclose (f);
    }
  else
    {
      printf ("Error : couldn't open '%s' for input.\n", argv[1]);
      exit (0);
    }

  if (stricmp(argv[2], "nil"))
  {
    if (!(fi = fopen (argv[2], "r")))
      {
	printf ("Error : couldn't open '%s' for input.\n", argv[2]);
	exit (0);
      }
  } else
  {
    fi = NULL;
    noin = 1;
  }

  if (!(fo = fopen (argv[3], "w")))
    {
      printf ("Error : couldn't open '%s' for output.\n", argv[3]);
      exit (0);
    }

  if (!(fs = fopen (argv[4], "w")))
    {
      printf ("Error : couldn't open '%s' for output.\n", argv[4]);
      exit (0);
    }

  clearConsole();

  #ifdef INTERACTIVE
  setupInteractiveConsole();
  #endif

  setConsolePosition(0, 0);

  for(y = 0; y < PAGEHEIGHT; y++)
  {
    for(x = 0; x < LINEWIDTH; x++)
    {
      if (isprint(cur)) printf("%c", cur);
    }
    printf("\n");
  }

  x = y = 0;
  oldX = oldY = 0;

  while ((cur != '@') || (stringmode))         /*** Intepreting Phase */
    {
      #ifdef INTERACTIVE

      if (!go)
      {
            /* Low-light the old PC (if it is a breakpoint, set to breakpoint color)*/
            if (breakPointActive && oldX == breakX && oldY == breakY)
            {
                debugHighlightChar(oldX, oldY, BREAKPOINT_COLOR);
            }
            else
            {
                debugHighlightChar(oldX, oldY, NORMAL_COLOR);
            }

            /* Highlight the current PC */
            debugHighlightChar(x, y, PC_COLOR);
      }
      oldX = x;
      oldY = y;

      /* Check for breakpoint in all modes. */
      if (breakPointActive && x == breakX && y == breakY)
      {
            debuggerSay("Breakpoint hit.");
            debugInput(x, y);
      }
      else
      {
          /* If we aren't animating and we aren't going, talk to the user. */
          if (!animate && !go)
          {
              debugInput(x, y);
          }
          /* We're animating or going; did the user break with the space key?*/
          else if (_kbhit() && (_getch() == ' '))
          {
               /* They broke; drop to debugger. */
               animate = go = 0;
               debugHighlightChar(x, y, PC_COLOR);
               debugInput(x, y);
          }
      }
      #else
      setConsolePosition(x, y);
      #endif

#ifndef WIN32
      fflush (stdout);
#endif
#ifdef DELAY
#ifdef INTERACTIVE
      if (animate)
      {
#endif
      DELAY(DELAY_MS);
      printPC(x, y);
#ifdef INTERACTIVE
      }
#endif
#endif
      if (stringmode && (cur != '"'))
	push (cur);
      else if (isdigit (cur))
	push (cur - '0');
      else
	switch (cur)
	   {
	   case '>':            /* PC Right */
	     dx = 1;
	     dy = 0;
	     break;
	   case '<':            /* PC Left */
	     dx = -1;
	     dy = 0;
	     break;
	   case '^':            /* PC Up */
	     dx = 0;
	     dy = -1;
	     break;
	   case 'v':            /* PC Down */
	     dx = 0;
	     dy = 1;
	     break;
	   case '|':            /* Vertical 'If' */
	     dx = 0;
	     if (pop ())
	       dy = -1;
	     else
	       dy = 1;
	     break;
	   case '_':            /* Horizontal 'If' */
	     dy = 0;
	     if (pop ())
	       dx = -1;
	     else
	       dx = 1;
	     break;
	   case '+':            /* Add */
	     push (pop () + pop ());
	     break;
	   case '-':            /* Subtract */
	     {
	       long a = pop();
	       long b = pop();
	       push(b - a);
	     }
	     break;
	   case '*':            /* Multiply */
	     push (pop () * pop ());
	     break;
	   case '/':            /* Integer Divide */
	     {
	       signed long a = pop ();
	       signed long b = pop ();
	       push (b / a);
	     }
	     break;
	   case '%':            /* Modulo */
	     {
	       signed long a = pop ();
	       signed long b = pop ();
	       push (b % a);
	     }
	     break;
	   case '\\':           /* Swap */
	     {
	       signed long a = pop ();
	       signed long b = pop ();
	       push (a);
	       push (b);
	     }
	     break;
	   case '.':            /* Pop Out Integer */
	     fprintf (fo, "%ld ", pop ());
	     fflush (fo);
	     break;
	   case ',':            /* Pop Out ASCII */
	     fprintf (fo, "%c", pop ());
	     fflush (fo);
	     break;
	   case '"':		/* Toggle String Mode */
	     stringmode = !stringmode;
	     break;
	   case ':':            /* Duplicate */
	     {
	       signed long a = pop ();
	       push (a);
	       push (a);
	     }
	     break;
	   case '!':            /* Negate */
	  if (pop())
	    push(0);
	  else
	    push(1);
	     break;
	   case '`':
	     {
	       signed long b = pop ();
	       signed long a = pop ();
	       push (a > b);
	     }
	     break;
	   case '#':            /* Bridge */
	     x += dx;
	     y += dy;
	     break;
	   case '$':            /* Pop and Discard */
	     pop ();
	     break;
	   case '?':            /* Random Redirect */
	     switch ((rand () / 32) % 4)
		{
		case 0:
		  dx = 1;
		  dy = 0;
		  break;
		case 1:
		  dx = -1;
		  dy = 0;
		  break;
		case 2:
		  dx = 0;
		  dy = -1;
		  break;
		case 3:
		  dx = 0;
		  dy = 1;
		  break;
		}
	     break;
	   case '&':            /* Input Integer */
	     {
	       signed long b;
	       if (!noin)
	       {
		 fscanf (fi, "%ld", &b);
		 push (b);
	       } else push (0);
	     }
	     break;
	   case '~':            /* Input ASCII */
	     {
	       char c;
	       if (!noin) c = fgetc (fi); else c = 0;
	       push (c);
	     }
	     break;
	   case 'g':            /* Get Value */
	     {
	       int y = (int)pop (), x = (int)pop ();
	       push (cur);
	     }
	     break;
	   case 'p':            /* Put Value */
	     {
	       int y = (int)pop (), x = (int)pop ();
	       cur = (char)pop ();
           setConsolePosition(x, y);
	       if (isprint (cur)) printf ("%c", cur); else printf(".");
	     }
	     break;
	   default:
	     break;
	   }
      x += dx;
      y += dy;
      if (x < 0)
	x = LINEWIDTH - 1; /* as pointed out by Matthew D Moss. */
			   /* backwards compatability for the  */
			   /* off-by-one error is only available */
			   /* in the interpreter... sorry! */
      else
	x = x % LINEWIDTH;
      if (y < 0)
	y = PAGEHEIGHT - 1; /* see above */
      else
	y = y % PAGEHEIGHT;
      {
	struct stack *s;
	for (s = head; s; s = s->next)
	  fprintf(fs, "%ld ", s->val);
	fprintf(fs, "\n");
	fflush(fs);
      }
    }

#ifdef INTERACTIVE
    debuggerSay("Program terminated.");
#endif

  exit (0);
  if (fi) fclose (fi);
  fclose (fo);
  fclose (fs);
  setConsolePosition(1, 22);
}

/*
 * pushes a value onto the stack.
 */
void push (signed long val)
{
  struct stack *s;
  s = (struct stack *) malloc (sizeof (struct stack));
  s->val = val;
  s->next = head;
  head = s;

  #ifdef INTERACTIVE
  if (!go)
  {
      printStack();
  }
  #endif
}

/*
 * pops a value off the stack. returns 0 in case of underflow.
 */
signed long pop ()
{
  signed long v;
  struct stack *s = head;
  if (s)
    {
      v = head->val;
      head = head->next;
      free (s);

      #ifdef INTERACTIVE
      if (!go)
      {
         printStack();
      }
      #endif

      return v;
    }
  else
    {
      return 0;
    }
}

void setConsolePosition(int x, int y)
{
#ifdef WIN32
    COORD   c;
    c.X = x;
    c.Y = y;
    SetConsoleCursorPosition(outputConsole, c);
#else
  printf ("%c[%d;%dH", 27, y + 1, x + 1);
#endif
}

void clearConsole(void)
{
#ifdef WIN32
    COORD c;
    CONSOLE_SCREEN_BUFFER_INFO screenInfo;
    DWORD junk;

    c.X = 0;
    c.Y = 0;

    if (GetConsoleScreenBufferInfo(outputConsole, &screenInfo))
    {
        FillConsoleOutputCharacter(outputConsole, ' ',
                      (DWORD)screenInfo.dwSize.X * (DWORD)screenInfo.dwSize.Y,
                      c, &junk);
    }                 
#else
printf ("%c[2J", 27);
#endif
}

#ifdef INTERACTIVE
#define MAX_DEB_INPUT       40

#define DEBUG_LINES         5
#define DEBUG_INPUT_SIZE    40

void setupInteractiveConsole(void)
{
    COORD c;
    int   i;

    c.X = LINEWIDTH + 1;
    c.Y = PAGEHEIGHT + DEBUG_LINES + 1;
    SetConsoleScreenBufferSize(outputConsole, c);
    setConsolePosition(0, PAGEHEIGHT);
    for (i = 0; i < LINEWIDTH + 1; i++)
    {
        putchar('-');
    }
    for (i = PAGEHEIGHT + 1; i <= PAGEHEIGHT + DEBUG_LINES; i++)
    {
        setConsolePosition(DEBUG_INPUT_SIZE + 1, i);
        putchar('|');
    }
    setConsolePosition(DEBUG_INPUT_SIZE + 3, PAGEHEIGHT);
    printf("Stack");
    setConsolePosition(1, PAGEHEIGHT);
    printf("Input");
}

void printStack(void)
{
    struct stack *s = head;
    int i;

    for (i = 1; i <= DEBUG_LINES; i++)
    {
        setConsolePosition(DEBUG_INPUT_SIZE + 3, PAGEHEIGHT + i);
        if (s)
        {
            printf("%d: %d          ", i, s->val) ;
            s = s->next;
        }
        else
        {
            if (i == 1)
            {
                printf("(empty stack)   ");
            }
            else
            {
                printf("                ");
            }
        }
    }
}

void scrollDebugWindow(void)
{
    SMALL_RECT scrollRect, clipRect;
    CHAR_INFO  fillChar;
    COORD      c;

    scrollRect.Left = 0;
    scrollRect.Top = PAGEHEIGHT + 2;
    scrollRect.Right = MAX_DEB_INPUT;
    scrollRect.Bottom = PAGEHEIGHT + DEBUG_LINES;

    memcpy(&clipRect, &scrollRect, sizeof(SMALL_RECT));
    clipRect.Top = PAGEHEIGHT + 1;

    fillChar.Char.AsciiChar = ' ';
    fillChar.Attributes = FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE;

    c.X = 0;
    c.Y = PAGEHEIGHT + 1;

    ScrollConsoleScreenBuffer(outputConsole, &scrollRect, &clipRect,
                              c, &fillChar);
}

void getLine(char *buffer, int maxSize)
{
    int i, c;

    /* ASSERT(maxSize > 1); */
    i = 0;

    while (i < (maxSize - 1))
    {
        c = _getch();
        if (c == '\n' || c == '\r')
        {
            break;
        }

        if (c == '\b')
        {
            if (i > 0)
            {
                i--;
                putchar(c);
                putchar(' ');
                putchar(c);
            }
        }
        else
        {
            buffer[i++] = c;
            putchar(c);
        }
    }

    buffer[i] = 0;
}

void debuggerSay(char *msg)
{
     scrollDebugWindow();
     setConsolePosition(0, PAGEHEIGHT + DEBUG_LINES);
     printf(msg);
     fflush(stdout);
}

void debugHighlightChar(int x, int y, WORD attribute)
{
    COORD c;
    DWORD junk;
    c.X = x;
    c.Y = y;
    WriteConsoleOutputAttribute(outputConsole, &attribute, 1, c, &junk);
}

void printPC(int x, int y)
{
    setConsolePosition(DEBUG_INPUT_SIZE + 20, PAGEHEIGHT + 1);
    printf("PC: (%d:%d)  ", x, y);
}

void debugInput(int x, int y)
{
    static char inputLine[MAX_DEB_INPUT];
    long        temp1;

    printPC(x, y);
    printStack();

    while (1)
    {
        scrollDebugWindow();
        setConsolePosition(0, PAGEHEIGHT + DEBUG_LINES);
        printf(">");
        getLine(inputLine, MAX_DEB_INPUT);
        switch (inputLine[0])
        {
            case 'a':
                     go = 0;
                     animate = 1;
                     debuggerSay("Animating, space to break...");
                     return;

            case 'g':
                     go = 1;
                     animate = 0;
                     debugHighlightChar(x, y, NORMAL_COLOR);
                     debuggerSay("Running, space to break...");
                     return;

            case 's':
                     go = animate = 0;
                     return;

            case 'q':
                     debuggerSay("Bye.");
                     exit(0);
                     break;

            case 'b':
                    if (breakPointActive)
                    {
                        debugHighlightChar(breakX, breakY, NORMAL_COLOR);
                        debuggerSay("Breakpoint cleared.");
                        breakPointActive = 0;
                    }
                    else
                    {
                        if (sscanf(inputLine, "b %d:%d", &breakX, &breakY) != 2)
                        {
                            debuggerSay("Heh? (syntax: b x:y)");
                        }
                        else
                        {
                            if (breakX < LINEWIDTH && breakY < PAGEHEIGHT)
                            {
                                debugHighlightChar(breakX, breakY, BREAKPOINT_COLOR);
                                breakPointActive = 1;
                            }
                            else
                            {
                                debuggerSay("Invalid breakpoint address.");
                            }
                        }
                    }
                    break;

            case '+':
                    if (sscanf(inputLine, "+ %ld", &temp1) != 1)
                    {
                        debuggerSay("Heh? (syntax + value)");
                    }
                    else
                    {
                        push(temp1);
                        printStack();
                        sprintf(inputLine, "Pushed %ld", temp1);
                        debuggerSay(inputLine);
                    }
                    break;

            case '-':
                   temp1 = pop();
                   printStack();
                   sprintf(inputLine, "Popped %ld", temp1);
                   debuggerSay(inputLine);
                   break;

            case '?':
                    debuggerSay("Commands: a (animate) g (go)");
                    debuggerSay(" s (single step) b (breakpoint)");
                    debuggerSay(" + (push) - (pop) q (quit)");
                    break;

            default:

                  debuggerSay("*** unknown command ***");
                  break;
        }
    }
}
#endif
